@@ -18,12 +18,10 @@ local rspamd_logger = require "rspamd_logger"
 local lua_util = require "lua_util"
 local dkim_sign_tools = require "lua_dkim_tools"
 local rspamd_util = require "rspamd_util"
-local rspamd_rsa_privkey = require "rspamd_rsa_privkey"
-local rspamd_rsa = require "rspamd_rsa"
 local fun = require "fun"
 local lua_auth_results = require "lua_auth_results"
-local hash = require "rspamd_cryptobox_hash"
 local lua_mime = require "lua_mime"
+local hash = require "rspamd_cryptobox_hash"
 
 if confighelp then
   return
@@ -39,6 +37,9 @@ end
 
 local dkim_verify = rspamd_plugins.dkim.verify
 local dkim_sign = rspamd_plugins.dkim.sign
+local dkim_load_sign_key = rspamd_plugins.dkim.load_sign_key
+local dkim_sign_key_get_alg = rspamd_plugins.dkim.sign_key_get_alg
+local dkim_sign_digest = rspamd_plugins.dkim.sign_digest
 local dkim_canonicalize = rspamd_plugins.dkim.canon_header_relaxed
 local redis_params
 
@@ -72,12 +73,13 @@ local settings = {
   use_domain = 'header',
   use_esld = true,
   use_redis = false,
-  key_prefix = 'arc_keys', -- default hash name
-  reuse_auth_results = false, -- Reuse the existing authentication results
+  key_prefix = 'arc_keys',       -- default hash name
+  reuse_auth_results = false,    -- Reuse the existing authentication results
   whitelisted_signers_map = nil, -- Trusted signers domains
-  adjust_dmarc = true, -- Adjust DMARC rejected policy for trusted forwarders
-  allowed_ids = nil, -- Allowed settings id
-  forbidden_ids = nil, -- Banned settings id
+  whitelist = nil,               -- Domains with broken ARC implementations to trust despite validation failures
+  adjust_dmarc = true,           -- Adjust DMARC rejected policy for trusted forwarders
+  allowed_ids = nil,             -- Allowed settings id
+  forbidden_ids = nil,           -- Banned settings id
 }
 
 -- To match normal AR
@@ -86,15 +88,15 @@ local ar_settings = lua_auth_results.default_settings
 local function parse_arc_header(hdr, target, is_aar)
   -- Split elements by ';' and trim spaces
   local arr = fun.totable(fun.map(
-      function(val)
-        return fun.totable(fun.map(lua_util.rspamd_str_trim,
-            fun.filter(function(v)
-              return v and #v > 0
-            end,
-                lua_util.rspamd_str_split(val.decoded, ';')
-            )
-        ))
-      end, hdr
+    function(val)
+      return fun.totable(fun.map(lua_util.rspamd_str_trim,
+        fun.filter(function(v)
+            return v and #v > 0
+          end,
+          lua_util.rspamd_str_split(val.decoded, ';')
+        )
+      ))
+    end, hdr
   ))
 
   -- v[1] is the key and v[2] is the value
@@ -115,11 +117,11 @@ local function parse_arc_header(hdr, target, is_aar)
     if not is_aar then
       -- For normal ARC headers we split by kv pair, like k=v
       fun.each(function(v)
-        fill_arc_header_table(v, target[i])
-      end,
-          fun.map(function(elt)
-            return lua_util.rspamd_str_split(elt, '=')
-          end, elts)
+          fill_arc_header_table(v, target[i])
+        end,
+        fun.map(function(elt)
+          return lua_util.rspamd_str_split(elt, '=')
+        end, elts)
       )
     else
       -- For AAR we check special case of i=%d and pass everything else to
@@ -147,7 +149,7 @@ local function parse_arc_header(hdr, target, is_aar)
 
   -- sort by i= attribute
   table.sort(target, function(a, b)
-    return (a.i or 0) < (b.i or 0)
+    return (tonumber(a.i) or 0) < (tonumber(b.i) or 0)
   end)
 end
 
@@ -156,14 +158,14 @@ local function arc_validate_seals(task, seals, sigs, seal_headers, sig_headers)
   for i = 1, #seals do
     if (sigs[i].i or 0) ~= i then
       fail_reason = string.format('bad i for signature: %d, expected %d; d=%s',
-          sigs[i].i, i, sigs[i].d)
+        sigs[i].i, i, sigs[i].d)
       rspamd_logger.infox(task, fail_reason)
       task:insert_result(arc_symbols['invalid'], 1.0, fail_reason)
       return false, fail_reason
     end
     if (seals[i].i or 0) ~= i then
       fail_reason = string.format('bad i for seal: %d, expected %d; d=%s',
-          seals[i].i, i, seals[i].d)
+        seals[i].i, i, seals[i].d)
       rspamd_logger.infox(task, fail_reason)
       task:insert_result(arc_symbols['invalid'], 1.0, fail_reason)
       return false, fail_reason
@@ -199,7 +201,31 @@ local function arc_callback(task)
   local arc_seal_headers = task:get_header_full('ARC-Seal')
   local arc_ar_headers = task:get_header_full('ARC-Authentication-Results')
 
+  lua_util.debugm(N, task, 'ARC verification: found %s ARC-MS, %s ARC-Seal, %s ARC-AR headers',
+                  arc_sig_headers and #arc_sig_headers or 0,
+                  arc_seal_headers and #arc_seal_headers or 0,
+                  arc_ar_headers and #arc_ar_headers or 0)
+
+  if arc_sig_headers then
+    for i, hdr in ipairs(arc_sig_headers) do
+      lua_util.debugm(N, task, 'ARC-Message-Signature[%s]: %s', i, hdr.decoded)
+    end
+  end
+
+  if arc_seal_headers then
+    for i, hdr in ipairs(arc_seal_headers) do
+      lua_util.debugm(N, task, 'ARC-Seal[%s]: %s', i, hdr.decoded)
+    end
+  end
+
+  if arc_ar_headers then
+    for i, hdr in ipairs(arc_ar_headers) do
+      lua_util.debugm(N, task, 'ARC-Authentication-Results[%s]: %s', i, hdr.decoded)
+    end
+  end
+
   if not arc_sig_headers or not arc_seal_headers then
+    lua_util.debugm(N, task, 'ARC verification failed: missing required headers')
     task:insert_result(arc_symbols['na'], 1.0)
     return
   end
@@ -207,7 +233,7 @@ local function arc_callback(task)
   if #arc_sig_headers ~= #arc_seal_headers then
     -- We mandate that count of seals is equal to count of signatures
     rspamd_logger.infox(task, 'number of seals (%s) is not equal to number of signatures (%s)',
-        #arc_seal_headers, #arc_sig_headers)
+      #arc_seal_headers, #arc_sig_headers)
     task:insert_result(arc_symbols['invalid'], 1.0, 'invalid count of seals and signatures')
     return
   end
@@ -249,7 +275,7 @@ local function arc_callback(task)
 
   -- Now check sanity of what we have
   local valid, validation_error = arc_validate_seals(task, cbdata.seals, cbdata.sigs,
-      arc_seal_headers, arc_sig_headers)
+    arc_seal_headers, arc_sig_headers)
   if not valid then
     task:cache_set('arc-failure', validation_error)
     return
@@ -267,12 +293,20 @@ local function arc_callback(task)
   local function gen_arc_seal_cb(index, sig)
     return function(_, res, err, domain)
       lua_util.debugm(N, task, 'checked arc seal: %s(%s), %s processed',
-          res, err, index)
+        res, err, index)
 
       if not res then
-        cbdata.res = 'fail'
-        if err and domain then
-          table.insert(cbdata.errors, string.format('sig:%s:%s', domain, err))
+        -- Check if this domain is whitelisted for broken ARC implementations
+        if settings.whitelist and domain and settings.whitelist:get_key(domain) then
+          rspamd_logger.infox(task, 'ARC seal validation failed for whitelisted domain %s, treating as valid: %s',
+            domain, err)
+          lua_util.debugm(N, task, 'whitelisted domain %s ARC seal failure ignored', domain)
+          res = true -- Treat as valid to continue the chain
+        else
+          cbdata.res = 'fail'
+          if err and domain then
+            table.insert(cbdata.errors, string.format('sig:%s:%s', domain, err))
+          end
         end
       end
 
@@ -283,7 +317,7 @@ local function arc_callback(task)
           local cur_aar = cbdata.ars[index]
           if not cur_aar then
             rspamd_logger.warnx(task, "cannot find Arc-Authentication-Results for trusted " ..
-                "forwarder %s on i=%s", domain, cbdata.index)
+              "forwarder %s on i=%s", domain, cbdata.index)
           else
             task:cache_set(AR_TRUSTED_CACHE_KEY, cur_aar)
             local seen_dmarc
@@ -309,20 +343,20 @@ local function arc_callback(task)
             end
           end
           task:insert_result(arc_symbols.trusted_allow, mult,
-              string.format('%s:s=%s:i=%d', domain, sig.s, index))
+            string.format('%s:s=%s:i=%d', domain, sig.s, index))
         end
       end
 
       if index == #arc_sig_headers then
         if cbdata.res == 'success' then
           local arc_allow_result = string.format('%s:s=%s:i=%d',
-              domain, sig.s, index)
+            domain, sig.s, index)
           task:insert_result(arc_symbols.allow, 1.0, arc_allow_result)
           task:cache_set('arc-allow', arc_allow_result)
         else
           task:insert_result(arc_symbols.reject, 1.0,
-              rspamd_logger.slog('seal check failed: %s, %s', cbdata.res,
-                  cbdata.errors))
+            rspamd_logger.slog('seal check failed: %s, %s', cbdata.res,
+              cbdata.errors))
         end
       end
     end
@@ -330,12 +364,20 @@ local function arc_callback(task)
 
   local function arc_signature_cb(_, res, err, domain)
     lua_util.debugm(N, task, 'checked arc signature %s: %s(%s)',
-        domain, res, err)
+      domain, res, err)
 
     if not res then
-      cbdata.res = 'fail'
-      if err and domain then
-        table.insert(cbdata.errors, string.format('sig:%s:%s', domain, err))
+      -- Check if this domain is whitelisted for broken ARC implementations
+      if settings.whitelist and domain and settings.whitelist:get_key(domain) then
+        rspamd_logger.infox(task, 'ARC signature validation failed for whitelisted domain %s, treating as valid: %s',
+          domain, err)
+        lua_util.debugm(N, task, 'whitelisted domain %s ARC signature failure ignored', domain)
+        res = true -- Treat as valid to continue the chain
+      else
+        cbdata.res = 'fail'
+        if err and domain then
+          table.insert(cbdata.errors, string.format('sig:%s:%s', domain, err))
+        end
       end
     end
     if cbdata.res == 'success' then
@@ -343,17 +385,24 @@ local function arc_callback(task)
       for i, sig in ipairs(cbdata.seals) do
         local ret, lerr = dkim_verify(task, sig.header, gen_arc_seal_cb(i, sig), 'arc-seal')
         if not ret then
-          cbdata.res = 'fail'
-          table.insert(cbdata.errors, string.format('seal:%s:s=%s:i=%s:%s',
+          -- Check if this domain is whitelisted for broken ARC implementations
+          if settings.whitelist and sig.d and settings.whitelist:get_key(sig.d) then
+            rspamd_logger.infox(task, 'ARC seal dkim_verify failed for whitelisted domain %s, treating as valid: %s',
+              sig.d, lerr)
+            lua_util.debugm(N, task, 'whitelisted domain %s ARC seal dkim_verify failure ignored', sig.d)
+          else
+            cbdata.res = 'fail'
+            table.insert(cbdata.errors, string.format('seal:%s:s=%s:i=%s:%s',
               sig.d or '', sig.s or '', sig.i or '', lerr))
-          lua_util.debugm(N, task, 'checked arc seal %s: %s(%s), %s processed',
+            lua_util.debugm(N, task, 'checked arc seal %s: %s(%s), %s processed',
               sig.d, ret, lerr, i)
+          end
         end
       end
     else
       task:insert_result(arc_symbols['reject'], 1.0,
-          rspamd_logger.slog('signature check failed: %s, %s', cbdata.res,
-              cbdata.errors))
+        rspamd_logger.slog('signature check failed: %s, %s', cbdata.res,
+          cbdata.errors))
     end
   end
 
@@ -397,25 +446,33 @@ local function arc_callback(task)
        is "fail" and the algorithm stops here.
    9.  If the algorithm reaches this step, then the Chain Validation
        Status is "pass", and the algorithm is complete.
-  ]]--
+  ]] --
 
   local processed = 0
   local sig = cbdata.sigs[#cbdata.sigs] -- last AMS
   local ret, err = dkim_verify(task, sig.header, arc_signature_cb, 'arc-sign')
 
   if not ret then
-    cbdata.res = 'fail'
-    table.insert(cbdata.errors, string.format('sig:%s:%s', sig.d or '', err))
+    -- Check if this domain is whitelisted for broken ARC implementations
+    if settings.whitelist and sig.d and settings.whitelist:get_key(sig.d) then
+      rspamd_logger.infox(task, 'ARC signature dkim_verify failed for whitelisted domain %s, treating as valid: %s',
+        sig.d, err)
+      lua_util.debugm(N, task, 'whitelisted domain %s ARC signature dkim_verify failure ignored', sig.d)
+      processed = processed + 1
+    else
+      cbdata.res = 'fail'
+      table.insert(cbdata.errors, string.format('sig:%s:%s', sig.d or '', err))
+    end
   else
     processed = processed + 1
     lua_util.debugm(N, task, 'processed arc signature %s[%s]: %s(%s), %s total',
-        sig.d, sig.i, ret, err, #cbdata.seals)
+      sig.d, sig.i, ret, err, #cbdata.seals)
   end
 
   if processed == 0 then
     task:insert_result(arc_symbols['reject'], 1.0,
-        rspamd_logger.slog('cannot verify %s of %s signatures: %s',
-            #arc_sig_headers - processed, #arc_sig_headers, cbdata.errors))
+      rspamd_logger.slog('cannot verify %s of %s signatures: %s',
+        #arc_sig_headers - processed, #arc_sig_headers, cbdata.errors))
   end
 end
 
@@ -491,27 +548,8 @@ rspamd_config:register_dependency('ARC_CHECK', 'SPF_CHECK')
 rspamd_config:register_dependency('ARC_CHECK', 'DKIM_CHECK')
 
 local function arc_sign_seal(task, params, header)
-  local arc_sigs = task:cache_get('arc-sigs')
   local arc_seals = task:cache_get('arc-seals')
-  local arc_auth_results = task:cache_get('arc-authres')
   local cur_auth_results
-  local privkey
-
-  if params.rawkey then
-    -- Distinguish between pem and base64
-    if string.match(params.rawkey, '^-----BEGIN') then
-      privkey = rspamd_rsa_privkey.load_pem(params.rawkey)
-    else
-      privkey = rspamd_rsa_privkey.load_base64(params.rawkey)
-    end
-  elseif params.key then
-    privkey = rspamd_rsa_privkey.load_file(params.key)
-  end
-
-  if not privkey then
-    rspamd_logger.errx(task, 'cannot load private key for signing')
-    return
-  end
 
   if settings.reuse_auth_results then
     local ar_header = task:get_header('Authentication-Results')
@@ -527,61 +565,100 @@ local function arc_sign_seal(task, params, header)
     cur_auth_results = lua_auth_results.gen_auth_results(task, ar_settings) or ''
   end
 
-  local sha_ctx = hash.create_specific('sha256')
-
-  -- Update using previous seals + sigs + AAR
   local cur_idx = 1
   if arc_seals then
     cur_idx = #arc_seals + 1
-    -- We use the cached version per each ARC-* header field individually, already sorted by instance
-    -- value in ascending order
-    for i = 1, #arc_seals, 1 do
+  end
+
+  header = lua_util.fold_header_with_encoding(task,
+    'ARC-Message-Signature',
+    header,
+    { structured = true, encode = false })
+
+  cur_auth_results = string.format('i=%d; %s', cur_idx, cur_auth_results)
+  cur_auth_results = lua_util.fold_header_with_encoding(task,
+    'ARC-Authentication-Results',
+    cur_auth_results,
+    { stop_chars = ';', structured = true, encode = false })
+
+  -- Create ARC-Seal signature manually using SHA256 hash
+  -- We must canonicalize all ARC headers in order and sign them
+  local sha_ctx = hash.create_specific('sha256')
+
+  -- Canonicalize previous ARC sets if they exist (arc_seals already retrieved above)
+  local arc_sigs = task:cache_get('arc-sigs')
+  local arc_auth_results = task:cache_get('arc-authres')
+
+  if arc_seals then
+    for i = 1, #arc_seals do
       if arc_auth_results[i] then
-        local s = dkim_canonicalize('ARC-Authentication-Results',
-            arc_auth_results[i].raw_header)
+        local s = dkim_canonicalize('ARC-Authentication-Results', arc_auth_results[i].raw_header)
         sha_ctx:update(s)
-        lua_util.debugm(N, task, 'update signature with header: %s', s)
+        lua_util.debugm(N, task, 'canonicalized previous AAR[%d]: %s', i, s)
       end
       if arc_sigs[i] then
-        local s = dkim_canonicalize('ARC-Message-Signature',
-            arc_sigs[i].raw_header)
+        local s = dkim_canonicalize('ARC-Message-Signature', arc_sigs[i].raw_header)
         sha_ctx:update(s)
-        lua_util.debugm(N, task, 'update signature with header: %s', s)
+        lua_util.debugm(N, task, 'canonicalized previous AMS[%d]: %s', i, s)
       end
       if arc_seals[i] then
         local s = dkim_canonicalize('ARC-Seal', arc_seals[i].raw_header)
         sha_ctx:update(s)
-        lua_util.debugm(N, task, 'update signature with header: %s', s)
+        lua_util.debugm(N, task, 'canonicalized previous AS[%d]: %s', i, s)
       end
     end
   end
 
-  header = lua_util.fold_header(task,
-      'ARC-Message-Signature',
-      header)
-
-  cur_auth_results = string.format('i=%d; %s', cur_idx, cur_auth_results)
-  cur_auth_results = lua_util.fold_header(task,
-      'ARC-Authentication-Results',
-      cur_auth_results, ';')
-
-  local s = dkim_canonicalize('ARC-Authentication-Results',
-      cur_auth_results)
+  -- Canonicalize the new ARC-Authentication-Results header
+  local s = dkim_canonicalize('ARC-Authentication-Results', cur_auth_results)
   sha_ctx:update(s)
-  lua_util.debugm(N, task, 'update signature with header: %s', s)
+  lua_util.debugm(N, task, 'canonicalized new AAR: %s', s)
+
+  -- Canonicalize the new ARC-Message-Signature header
   s = dkim_canonicalize('ARC-Message-Signature', header)
   sha_ctx:update(s)
-  lua_util.debugm(N, task, 'update signature with header: %s', s)
+  lua_util.debugm(N, task, 'canonicalized new AMS: %s', s)
 
-  local cur_arc_seal = string.format('i=%d; s=%s; d=%s; t=%d; a=rsa-sha256; cv=%s; b=',
-      cur_idx,
-      params.selector,
-      params.domain,
-      math.floor(rspamd_util.get_time()), params.arc_cv)
-  s = string.format('%s:%s', 'arc-seal', cur_arc_seal)
-  sha_ctx:update(s)
-  lua_util.debugm(N, task, 'initial update signature with header: %s', s)
+  -- Load the signing key
+  local sign_key = dkim_load_sign_key(task, {
+    key = params.key,
+    rawkey = params.rawkey,
+  })
 
+  if not sign_key then
+    rspamd_logger.errx(task, 'cannot load signing key')
+    return
+  end
+
+  -- Get the algorithm from the key
+  local algorithm = dkim_sign_key_get_alg(sign_key)
+  if not algorithm or algorithm == 'unknown' then
+    rspamd_logger.errx(task, 'cannot determine key algorithm')
+    return
+  end
+
+  -- Construct partial ARC-Seal header (without signature) using detected algorithm
+  local cur_arc_seal = string.format('i=%d; a=%s; d=%s; s=%s; cv=%s; t=%d; b=',
+    cur_idx, algorithm, params.domain, params.selector, params.arc_cv,
+    math.floor(rspamd_util.get_time()))
+
+  -- Canonicalize the partial ARC-Seal and add to hash
+  local seal_canon = string.format('%s:%s', 'arc-seal', cur_arc_seal)
+  sha_ctx:update(seal_canon)
+  lua_util.debugm(N, task, 'canonicalized partial seal: %s', seal_canon)
+
+  -- Now sign the complete digest
+  local dret, sig_b64 = dkim_sign_digest(task, {
+    sign_key = sign_key,
+    digest = sha_ctx:bin(),
+  })
+
+  if not dret or not sig_b64 then
+    rspamd_logger.errx(task, 'cannot create ARC seal signature')
+    return
+  end
+
+  -- Get appropriate line ending
   local nl_type
   if task:has_flag("milter") then
     nl_type = "lf"
@@ -589,22 +666,31 @@ local function arc_sign_seal(task, params, header)
     nl_type = task:get_newlines_type()
   end
 
-  local sig = rspamd_rsa.sign_memory(privkey, sha_ctx:bin())
-  cur_arc_seal = string.format('%s%s', cur_arc_seal,
-      sig:base64(70, nl_type))
+  -- Fold the signature to 70 chars per line
+  local folded_sig = rspamd_util.encode_base64(rspamd_util.decode_base64(sig_b64), 70, nl_type)
+  cur_arc_seal = cur_arc_seal .. folded_sig
+
+  -- Add all ARC headers in a single call with explicit ordering
+  lua_util.debugm(N, task, 'adding ARC-Authentication-Results: %s', cur_auth_results)
+  lua_util.debugm(N, task, 'adding ARC-Message-Signature: %s', header)
+  lua_util.debugm(N, task, 'adding ARC-Seal: %s', cur_arc_seal)
 
   lua_mime.modify_headers(task, {
     add = {
       ['ARC-Authentication-Results'] = { order = 1, value = cur_auth_results },
       ['ARC-Message-Signature'] = { order = 1, value = header },
-      ['ARC-Seal'] = { order = 1, value = lua_util.fold_header(task,
-          'ARC-Seal', cur_arc_seal) }
+      ['ARC-Seal'] = {
+        order = 1,
+        value = lua_util.fold_header_with_encoding(task,
+          'ARC-Seal', cur_arc_seal,
+          { structured = true, encode = false })
+      }
     },
-    -- RFC requires a strict order for these headers to be inserted
+    -- RFC 8617 requires strict ordering of ARC headers
     order = { 'ARC-Authentication-Results', 'ARC-Message-Signature', 'ARC-Seal' },
   })
   task:insert_result(settings.sign_symbol, 1.0,
-      string.format('%s:s=%s:i=%d', params.domain, params.selector, cur_idx))
+    string.format('%s:s=%s:i=%d', params.domain, params.selector, cur_idx))
 end
 
 local function prepare_arc_selector(task, sel)
@@ -668,18 +754,12 @@ local function prepare_arc_selector(task, sel)
     else
       default_arc_cv()
     end
-
   end
 
   return true
 end
 
 local function do_sign(task, sign_params)
-  if sign_params.alg and sign_params.alg ~= 'rsa' then
-    -- No support for ed25519 keys
-    return
-  end
-
   if not prepare_arc_selector(task, sign_params) then
     -- Broken arc
     return
@@ -695,19 +775,18 @@ local function do_sign(task, sign_params)
           sign_params.pubkey = results[1]
           sign_params.strict_pubkey_check = not settings.allow_pubkey_mismatch
         elseif not settings.allow_pubkey_mismatch then
-          rspamd_logger.errx('public key for domain %s/%s is not found: %s, skip signing',
-              sign_params.domain, sign_params.selector, err)
+          rspamd_logger.errx(task, 'public key for domain %s/%s is not found: %s, skip signing',
+            sign_params.domain, sign_params.selector, err)
           return
         else
-          rspamd_logger.infox('public key for domain %s/%s is not found: %s',
-              sign_params.domain, sign_params.selector, err)
+          rspamd_logger.infox(task, 'public key for domain %s/%s is not found: %s',
+            sign_params.domain, sign_params.selector, err)
         end
 
         local dret, hdr = dkim_sign(task, sign_params)
         if dret then
           arc_sign_seal(task, sign_params, hdr)
         end
-
       end,
       forced = true
     })
@@ -768,6 +847,31 @@ end
 
 dkim_sign_tools.process_signing_settings(N, settings, opts)
 
+-- Process ARC-specific maps that aren't handled by dkim_sign_tools
+local lua_maps = require "lua_maps"
+
+if opts.whitelisted_signers_map then
+  settings.whitelisted_signers_map = lua_maps.map_add_from_ucl(opts.whitelisted_signers_map, 'set',
+    'ARC trusted signers domains')
+  if not settings.whitelisted_signers_map then
+    rspamd_logger.errx(rspamd_config, 'cannot load whitelisted_signers_map')
+    settings.whitelisted_signers_map = nil
+  else
+    rspamd_logger.infox(rspamd_config, 'loaded ARC whitelisted signers map')
+  end
+end
+
+if opts.whitelist then
+  settings.whitelist = lua_maps.map_add_from_ucl(opts.whitelist, 'set',
+    'ARC domains with broken implementations')
+  if not settings.whitelist then
+    rspamd_logger.errx(rspamd_config, 'cannot load ARC whitelist map')
+    settings.whitelist = nil
+  else
+    rspamd_logger.infox(rspamd_config, 'loaded ARC whitelist map')
+  end
+end
+
 if not dkim_sign_tools.validate_signing_settings(settings) then
   rspamd_logger.infox(rspamd_config, 'mandatory parameters missing, disable arc signing')
   return
@@ -780,7 +884,7 @@ if ar_opts and ar_opts.routines then
 
   if routines['authentication-results'] then
     ar_settings = lua_util.override_defaults(ar_settings,
-        routines['authentication-results'])
+      routines['authentication-results'])
   end
 end
 
@@ -789,7 +893,7 @@ if settings.use_redis then
 
   if not redis_params then
     rspamd_logger.errx(rspamd_config, 'no servers are specified, ' ..
-        'but module is configured to load keys from redis, disable arc signing')
+      'but module is configured to load keys from redis, disable arc signing')
     return
   end
 
@@ -845,9 +949,9 @@ if settings.adjust_dmarc and settings.whitelisted_signers_map then
           local dmarc_fwd = ar.dmarc
           if dmarc_fwd == 'pass' then
             rspamd_logger.infox(task, "adjust dmarc reject score as trusted forwarder "
-                .. "proved DMARC validity for %s", ar['header.from'])
+              .. "proved DMARC validity for %s", ar['header.from'])
             task:adjust_result(sym_to_adjust, 0.1,
-                'ARC trusted')
+              'ARC trusted')
           end
         end
       end
