@@ -14,10 +14,6 @@ See the License for the specific language governing permissions and
 limitations under the License.
 ]]--
 
-if confighelp then
-  return
-end
-
 -- A generic plugin for reputation handling
 
 local E = {}
@@ -32,7 +28,8 @@ local hash = require 'rspamd_cryptobox_hash'
 local lua_redis = require "lua_redis"
 local fun = require "fun"
 local lua_selectors = require "lua_selectors"
-local ts = require("tableshape").types
+local T = require "lua_shape.core"
+local PluginSchema = require "lua_shape.plugin_schema"
 
 local redis_params = nil
 local default_expiry = 864000 -- 10 day by default
@@ -200,7 +197,9 @@ local function dkim_reputation_filter(task, rule)
           end
         end
 
-        if sel_tld and requests[sel_tld] then
+        if rule.selector.config.exclusion_map and sel_tld and rule.selector.config.exclusion_map:get_key(sel_tld) then
+          lua_util.debugm(N, task, 'DKIM domain %s is excluded from reputation scoring', sel_tld)
+        elseif sel_tld and requests[sel_tld] then
           if requests[sel_tld] == 'a' then
             rep_accepted = rep_accepted + generic_reputation_calc(v, rule, 1.0, task)
           end
@@ -243,9 +242,13 @@ local function dkim_reputation_idempotent(task, rule)
 
   if sc then
     for dom, res in pairs(requests) do
-      -- tld + "." + check_result, e.g. example.com.+ - reputation for valid sigs
-      local query = string.format('%s.%s', dom, res)
-      rule.backend.set_token(task, rule, nil, query, sc)
+      if rule.selector.config.exclusion_map and rule.selector.config.exclusion_map:get_key(dom) then
+        lua_util.debugm(N, task, 'DKIM domain %s is excluded from reputation update', dom)
+      else
+        -- tld + "." + check_result, e.g. example.com.+ - reputation for valid sigs
+        local query = string.format('%s.%s', dom, res)
+        rule.backend.set_token(task, rule, nil, query, sc)
+      end
     end
   end
 end
@@ -277,6 +280,7 @@ local dkim_selector = {
     outbound = true,
     inbound = true,
     max_accept_adjustment = 2.0, -- How to adjust accepted DKIM score
+    exclusion_map = nil
   },
   dependencies = { "DKIM_TRACE" },
   filter = dkim_reputation_filter, -- used to get scores
@@ -356,10 +360,14 @@ local function url_reputation_filter(task, rule)
         for i, res in pairs(results) do
           local req = requests[i]
           if req then
-            local url_score = generic_reputation_calc(res, rule,
-                req[2] / mhits, task)
-            lua_util.debugm(N, task, "score for url %s is %s, score=%s", req[1], url_score, score)
-            score = score + url_score
+            if rule.selector.config.exclusion_map and rule.selector.config.exclusion_map:get_key(req[1]) then
+              lua_util.debugm(N, task, 'URL domain %s is excluded from reputation scoring', req[1])
+            else
+              local url_score = generic_reputation_calc(res, rule,
+                  req[2] / mhits, task)
+              lua_util.debugm(N, task, "score for url %s is %s, score=%s", req[1], url_score, score)
+              score = score + url_score
+            end
           end
         end
 
@@ -386,7 +394,11 @@ local function url_reputation_idempotent(task, rule)
 
   if sc then
     for _, tld in ipairs(requests) do
-      rule.backend.set_token(task, rule, nil, tld[1], sc)
+      if rule.selector.config.exclusion_map and rule.selector.config.exclusion_map:get_key(tld[1]) then
+        lua_util.debugm(N, task, 'URL domain %s is excluded from reputation update', tld[1])
+      else
+        rule.backend.set_token(task, rule, nil, tld[1], sc)
+      end
     end
   end
 end
@@ -401,6 +413,7 @@ local url_selector = {
     check_from = true,
     outbound = true,
     inbound = true,
+    exclusion_map = nil
   },
   filter = url_reputation_filter, -- used to get scores
   idempotent = url_reputation_idempotent -- used to set scores
@@ -439,6 +452,11 @@ local function ip_reputation_filter(task, rule)
     ip = ip:apply_mask(cfg.ipv6_mask)
   end
 
+  if cfg.exclusion_map and cfg.exclusion_map:get_key(ip) then
+    lua_util.debugm(N, task, 'IP %s is excluded from reputation scoring', tostring(ip))
+    return
+  end
+
   local pool = task:get_mempool()
   local asn = pool:get_variable("asn")
   local country = pool:get_variable("country")
@@ -554,6 +572,11 @@ local function ip_reputation_idempotent(task, rule)
     ip = ip:apply_mask(cfg.ipv6_mask)
   end
 
+  if cfg.exclusion_map and cfg.exclusion_map:get_key(ip) then
+    lua_util.debugm(N, task, 'IP %s is excluded from reputation update', tostring(ip))
+    return
+  end
+
   local pool = task:get_mempool()
   local asn = pool:get_variable("asn")
   local country = pool:get_variable("country")
@@ -600,6 +623,7 @@ local ip_selector = {
     inbound = true,
     ipv4_mask = 32, -- Mask bits for ipv4
     ipv6_mask = 64, -- Mask bits for ipv6
+    exclusion_map = nil
   },
   --dependencies = {"ASN"}, -- ASN is a prefilter now...
   init = ip_reputation_init,
@@ -621,6 +645,11 @@ local function spf_reputation_filter(task, rule)
   local cr = require "rspamd_cryptobox_hash"
   local hkey = cr.create(spf_record):base32():sub(1, 32)
 
+  if rule.selector.config.exclusion_map and rule.selector.config.exclusion_map:get_key(hkey) then
+    lua_util.debugm(N, task, 'SPF record %s is excluded from reputation scoring', hkey)
+    return
+  end
+
   lua_util.debugm(N, task, 'check spf record %s -> %s', spf_record, hkey)
 
   local function tokens_cb(err, token, values)
@@ -649,6 +678,11 @@ local function spf_reputation_idempotent(task, rule)
   local cr = require "rspamd_cryptobox_hash"
   local hkey = cr.create(spf_record):base32():sub(1, 32)
 
+  if rule.selector.config.exclusion_map and rule.selector.config.exclusion_map:get_key(hkey) then
+    lua_util.debugm(N, task, 'SPF record %s is excluded from reputation update', hkey)
+    return
+  end
+
   lua_util.debugm(N, task, 'set spf record %s -> %s = %s',
       spf_record, hkey, sc)
   rule.backend.set_token(task, rule, nil, hkey, sc)
@@ -663,6 +697,7 @@ local spf_selector = {
     max_score = nil,
     outbound = true,
     inbound = true,
+    exclusion_map = nil
   },
   dependencies = { "R_SPF_ALLOW" },
   filter = spf_reputation_filter, -- used to get scores
@@ -697,6 +732,13 @@ local function generic_reputation_init(rule)
         'Whitelisted selectors')
   end
 
+  if cfg.exclusion_map then
+    cfg.exclusion_map = lua_maps.map_add('reputation',
+        'generic_exclusion',
+        'set',
+        'Excluded selectors')
+  end
+
   return true
 end
 
@@ -706,6 +748,10 @@ local function generic_reputation_filter(task, rule)
 
   local function tokens_cb(err, token, values)
     if values then
+      if cfg.exclusion_map and cfg.exclusion_map:get_key(token) then
+        lua_util.debugm(N, task, 'Generic selector token %s is excluded from reputation scoring', token)
+        return
+      end
       local score = generic_reputation_calc(values, rule, 1.0, task)
 
       if math.abs(score) > 1e-3 then
@@ -742,29 +788,47 @@ local function generic_reputation_idempotent(task, rule)
   if sc then
     if type(selector_res) == 'table' then
       fun.each(function(e)
-        lua_util.debugm(N, task, 'set generic selector (%s) %s = %s',
-            rule['symbol'], e, sc)
-        rule.backend.set_token(task, rule, nil, e, sc)
+        if cfg.exclusion_map and cfg.exclusion_map:get_key(e) then
+          lua_util.debugm(N, task, 'Generic selector token %s is excluded from reputation update', e)
+        else
+          lua_util.debugm(N, task, 'set generic selector (%s) %s = %s',
+              rule['symbol'], e, sc)
+          rule.backend.set_token(task, rule, nil, e, sc)
+        end
       end, selector_res)
     else
-      lua_util.debugm(N, task, 'set generic selector (%s) %s = %s',
-          rule['symbol'], selector_res, sc)
-      rule.backend.set_token(task, rule, nil, selector_res, sc)
+      if cfg.exclusion_map and cfg.exclusion_map:get_key(selector_res) then
+        lua_util.debugm(N, task, 'Generic selector token %s is excluded from reputation update', selector_res)
+      else
+        lua_util.debugm(N, task, 'set generic selector (%s) %s = %s',
+            rule['symbol'], selector_res, sc)
+        rule.backend.set_token(task, rule, nil, selector_res, sc)
+      end
     end
   end
 end
 
 local generic_selector = {
-  schema = ts.shape {
-    lower_bound = ts.number + ts.string / tonumber,
-    max_score = ts.number:is_optional(),
-    min_score = ts.number:is_optional(),
-    outbound = ts.boolean,
-    inbound = ts.boolean,
-    selector = ts.string,
-    delimiter = ts.string,
-    whitelist = ts.one_of(lua_maps.map_schema, lua_maps_exprs.schema):is_optional(),
-  },
+  schema = T.table({
+    lower_bound = T.one_of({
+      T.number(),
+      T.transform(T.string(), tonumber)
+    }):doc({ summary = "Minimum number of messages to be scored" }),
+    max_score = T.number():optional():doc({ summary = "Maximum score" }),
+    min_score = T.number():optional():doc({ summary = "Minimum score" }),
+    outbound = T.boolean():doc({ summary = "Apply to outbound messages" }),
+    inbound = T.boolean():doc({ summary = "Apply to inbound messages" }),
+    selector = T.string():doc({ summary = "Selector expression" }),
+    delimiter = T.string():doc({ summary = "Selector delimiter" }),
+    whitelist = T.one_of({
+      { name = "map", schema = lua_maps.map_schema },
+      { name = "expr", schema = lua_maps_exprs.schema }
+    }):optional():doc({ summary = "Whitelist map or expression" }),
+    exclusion_map = T.one_of({
+      { name = "map", schema = lua_maps.map_schema },
+      { name = "expr", schema = lua_maps_exprs.schema }
+    }):optional():doc({ summary = "Exclusion map or expression" })
+  }):doc({ summary = "Generic selector reputation backend configuration" }),
   config = {
     lower_bound = 10, -- minimum number of messages to be scored
     min_score = nil,
@@ -773,13 +837,16 @@ local generic_selector = {
     inbound = true,
     selector = nil,
     delimiter = ':',
-    whitelist = nil
+    whitelist = nil,
+    exclusion_map = nil
   },
   init = generic_reputation_init,
   filter = generic_reputation_filter, -- used to get scores
   idempotent = generic_reputation_idempotent -- used to set scores
 }
 
+PluginSchema.register("plugins.reputation.selector.generic", generic_selector.schema)
+
 local selectors = {
   ip = ip_selector,
   sender = ip_selector, -- Better name
@@ -1090,13 +1157,22 @@ end
 local backends = {
   redis = {
     schema = lua_redis.enrich_schema({
-      prefix = ts.string:is_optional(),
-      expiry = (ts.number + ts.string / lua_util.parse_time_interval):is_optional(),
-      buckets = ts.array_of(ts.shape {
-        time = ts.number + ts.string / lua_util.parse_time_interval,
-        name = ts.string,
-        mult = ts.number + ts.string / tonumber
-      })          :is_optional(),
+      prefix = T.string():optional():doc({ summary = "Redis key prefix" }),
+      expiry = T.one_of({
+        T.number(),
+        T.transform(T.string(), lua_util.parse_time_interval)
+      }):optional():doc({ summary = "Default expiry time (seconds)" }),
+      buckets = T.array(T.table({
+        time = T.one_of({
+          T.number(),
+          T.transform(T.string(), lua_util.parse_time_interval)
+        }):doc({ summary = "Bucket time window (seconds)" }),
+        name = T.string():doc({ summary = "Bucket name" }),
+        mult = T.one_of({
+          T.number(),
+          T.transform(T.string(), tonumber)
+        }):doc({ summary = "Bucket multiplier" })
+      })):optional():doc({ summary = "Time buckets configuration" }),
     }),
     config = {
       expiry = default_expiry,
@@ -1107,16 +1183,16 @@ local backends = {
           name = '1m',
           mult = 1.0,
         }
-      }, -- What buckets should be used, default 1h and 1month
+      }, -- What buckets should be used, default 1month
     },
     init = reputation_redis_init,
     get_token = reputation_redis_get_token,
     set_token = reputation_redis_set_token,
   },
   dns = {
-    schema = ts.shape {
-      list = ts.string,
-    },
+    schema = T.table({
+      list = T.string():doc({ summary = "DNS list domain" }),
+    }):doc({ summary = "DNS reputation backend configuration" }),
     config = {
       -- list = rep.example.com
     },
@@ -1126,6 +1202,13 @@ local backends = {
   }
 }
 
+PluginSchema.register("plugins.reputation.backend.redis", backends.redis.schema)
+PluginSchema.register("plugins.reputation.backend.dns", backends.dns.schema)
+
+if confighelp then
+  return
+end
+
 local function is_rule_applicable(task, rule)
   local ip = task:get_from_ip()
   if not (rule.selector.config.outbound and rule.selector.config.inbound) then
@@ -1233,10 +1316,10 @@ local function parse_rule(name, tbl)
   rule.config = lua_util.override_defaults(rule.config, tbl)
 
   if rule.config.whitelist then
-    if lua_maps_exprs.schema(rule.config.whitelist) then
+    if lua_maps_exprs.schema:check(rule.config.whitelist) then
       rule.config.whitelist_map = lua_maps_exprs.create(rspamd_config,
           rule.config.whitelist, N)
-    elseif lua_maps.map_schema(rule.config.whitelist) then
+    elseif lua_maps.map_schema:check(rule.config.whitelist) then
       local map = lua_maps.map_add_from_ucl(rule.config.whitelist,
           'radix',
           sel_type .. ' reputation whitelist')
@@ -1267,6 +1350,24 @@ local function parse_rule(name, tbl)
     end
   end
 
+  -- Parse exclusion_map for reputation exclusion lists
+  if rule.config.exclusion_map then
+    local map_type = 'set' -- Default to set for string-based selectors (dkim, url, spf, generic)
+    if sel_type == 'ip' or sel_type == 'sender' then
+      map_type = 'radix' -- Use radix for IP-based selectors
+    end
+    local map = lua_maps.map_add_from_ucl(rule.config.exclusion_map,
+        map_type,
+        sel_type .. ' reputation exclusion map')
+    if not map then
+      rspamd_logger.errx(rspamd_config, "cannot parse exclusion map config for %s: (%s)",
+          sel_type,
+          rule.config.exclusion_map)
+      return false
+    end
+    rule.config.exclusion_map = map
+  end
+
   local symbol = rule.selector.config.symbol or name
   if tbl.symbol then
     symbol = tbl.symbol
@@ -1387,4 +1488,4 @@ if opts['rules'] then
   end
 else
   lua_util.disable_module(N, "config")
-end
+end
\ No newline at end of file
